/****************************************************************
 *  SFC.C                                                       *
 *              Dateivergleich (Smart File-Compare)             *
 *              ===================================             *
 *                                                              *
 *  30.07.91    Bodo Wenzel     V1.0, Erstprogrammierung        *
 *  21.06.94    Bodo Wenzel     V1.1, verbesserte Bedienung     *
 *  23.07.98    Bodo Wenzel     V1.2, Suchen mehrere Bytes      *
 *                              Offset direkt, Zeilenenden      *
 *  19.08.98    Bodo Wenzel     V2.0, auch fr TOS              *
 ****************************************************************/

#define VERSION "V2.0"

#include <ctype.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>

/* --- Konstanten und Makros ---------------------------------- */

#define MAXSTR          (FILENAME_MAX+1)

#define ERR_NONE        (0)
#define ERR_ARGS        (1)
#define ERR_NOTEQUAL    (2)
#define ERR_NOFILE      (3)
#define ERR_EMPTY       (4)

#define FALSE           (0)
#define TRUE            (!FALSE)

#define KEY_SLICE       (1234)

/* --- Globale Variablen -------------------------------------- */

char    name1[MAXSTR],name2[MAXSTR];
FILE    *file1,*file2;
long    length1,length2;
long    offset1,offset2;
short   which;
char    mode;
char    search_jump;
short   search_n,search_byte[4];
short   jump_byte;

/* --- Funktionsprototypen ------------------------------------ */

void usage(char *name);
short open_files(void);
void close_files(void);
long file_compare(void);

void show_files(void);
void init_display(void);
void exit_display(void);
void display_title(void);
void display_title1(long offset);
void display1(void);
void display_title2(long offset);
void display2(void);
void display_help(void);

long get_offset(void);
short get_search(void);
short get_jump(void);

void compare(void);
void search(void);
void jump(void);

void set_relative(short who,long difference);
void set_absolute(short who,long offset);
short frgetc(FILE *file);
void set_relative_up(short who);
void set_relative_down(short who);
void set_relative_pgup(short who);
void set_relative_pgdown(short who);

/* --- DOS-Schnittstelle -------------------------------------- */

#ifdef MSDOS

#include <ctutl.h>
#include <ctvio.h>

#define ASCII_NONE      (0x07)
#define LF              (0x0A)

#define sfc_kbhit()             UtlTaste()
#define sfc_getkey()            UtlTaste()
#define sfc_kflags()            UtlKbFlags()

void sfc_wait_nokey(void) {
/*-----------------------*/
while (UtlTaste()!=0);
}

void sfc_cls(void) {
/*----------------*/
VioSpza(' ',NORMAL);
VioSwa(0,1,80,1,INVERS);
VioSwa(0,13,80,1,INVERS);
}

#define sfc_goto_xy(x,y)        VioSetCursorPos(x,y)
#define sfc_line(y,s)           (VioSwz(0,y,80,1,' '),          \
				 VioSs(0,y,s))
#define sfc_invers(x,y,s)       VioSs(x,y,s)
#define sfc_normal(x,y,s)       VioSs(x,y,s)

short   buffer[16*60];

void sfc_help_save(void) {
/*----------------------*/
VioSw2b(10,5,60,16,(WORD far *)buffer);
UtlZeigeRahmen(10,5,60,16,RT_EDDE,HELL);
VioSwza(11,6,58,14,' ',NORMAL);
}

void sfc_help_restore(void) {
/*-------------------------*/
VioSb2w(10,5,60,16,(WORD far *)buffer);
if (which==1)
  sfc_goto_xy(11,1);
else
  sfc_goto_xy(11,13);
}

#endif

/* --- TOS-Schnittstelle -------------------------------------- */

#ifdef __TOS__

#include <tos.h>

#define ASCII_NONE      (0x03)
#define LF              (0x0A)

#define T_ESC           (0x01B)
#define T_RETURN        (0x00D)
#define T_BACKSPACE     (0x008)
#define T_LEFT          (0x14B)
#define T_RIGHT         (0x14D)
#define T_UP            (0x148)
#define T_DOWN          (0x150)
#define T_PGUP          (0x248)
#define T_PGDN          (0x250)
#define T_HOME          (0x273)
#define T_END           (0x274)

int sfc_kbhit(void) {
/*-----------------*/
if (!Cconis())
  return(FALSE);
(void)Crawcin();
return(TRUE);
}

#define sfc_wait_nokey()

int sfc_getkey(void) {
/*------------------*/
long    k;

k=Crawcin();

switch ((int)(k>>16)&0xFF) {
case 0x48:
case 0x4B:
case 0x4D:
case 0x50:
case 0x73:
case 0x74:
  k&=~0xFF;
  break;
default:
  break;
  }

if ((k&0xFF)==0) {
  k=0x100+((k>>16)&0xFF);
  if (Kbshift(-1)&4)
    k+=0x100;
} else
  k&=0xFF;

return((int)k);
}

#define RIGHT_SHIFT     (1)
#define LEFT_SHIFT      (2)

#define sfc_kflags()    ((int)Kbshift(-1))

void sfc_cls(void) {
/*----------------*/
printf("\033E\033w"
       "\033Y! \033p%80s"
       "\033Y- %80s\033q","","");
Cursconf(1,0);
Cursconf(2,0);
}

#define sfc_goto_xy(x,y)        printf("\033Y%c%c",' '+y,' '+x)
#define sfc_line(y,s)           printf("\033Y%c \033l%s",' '+y,s)
#define sfc_invers(x,y,s)       printf("\033j"                  \
				       "\033p\033Y%c%c%s\033q"  \
				       "\033k",' '+y,' '+x,s)

void sfc_normal(int x,int y,char *s) {
/*----------------------------------*/
printf("\033j\033Y%c%c",' '+y,' '+x);
while (*s)
  Bconout(5,*s++);
printf("\033k");
}

void sfc_help_save(void) {
/*----------------------*/
int     i,j;

printf("\033Y%%*");
for (i=0; i<60; i++)
  putchar('#');
for (i=0; i<14; i++) {
  printf("\033Y%c*",i+'&');
  putchar('#');
  for (j=0; j<58; j++)
    putchar(' ');
  putchar('#');
  }
printf("\033Y4*");
for (i=0; i<60; i++)
  putchar('#');
}

#define sfc_help_restore()      (sfc_cls(),                     \
				 display_title1(offset1),       \
				 display1(),                    \
				 display_title2(offset2),       \
				 display2(),                    \
				 display_title())

#endif

/* --- M A I N ------------------------------------------------ */

short main(short argc,char *argv[]) {
/*---------------------------------*/
short   ret;
long    offset;

if (argv[0][0]=='\0')
  argv[0]="SFC";

if (argc==3) {
  strcpy(name1,argv[1]);
  strcpy(name2,argv[2]);
  ret=open_files();
  if (ret==ERR_NONE) {
    offset=file_compare();
    if (offset>=0L) {
      offset1=offset;
      offset2=offset;
      show_files();
      ret=ERR_NOTEQUAL;
      }
    close_files();
  } else
    usage(argv[0]);
} else {
  usage(argv[0]);
  ret=ERR_ARGS;
  }
return(ret);
}

/* --- Hauptfunktionen ---------------------------------------- */

void usage(char *name) {
/*--------------------*/
printf("Smart FileCompare " VERSION " (" __DATE__ "), usage:\n"
       "%s file1 file2\n",name);
}

short open_files(void) {
/*--------------------*/
short   ret;

file1=fopen(name1,"rb");
if (file1==NULL) {
  ret=ERR_NOFILE;
} else {
  fseek(file1,0L,SEEK_END);
  length1=ftell(file1);
  if (length1==0L) {
    fclose(file1);
    ret=ERR_EMPTY;
  } else {
    file2=fopen(name2,"rb");
    if (file2==NULL) {
      fclose(file1);
      ret=ERR_NOFILE;
    } else {
      fseek(file2,0L,SEEK_END);
      length2=ftell(file2);
      if (length2==0L) {
	fclose(file1);
	fclose(file2);
	ret=ERR_EMPTY;
      } else {
	fseek(file1,0L,SEEK_SET);
	fseek(file2,0L,SEEK_SET);
	ret=ERR_NONE;
	}
      }
    }
  }
return(ret);
}

void close_files(void) {
/*--------------------*/
fclose(file1);
fclose(file2);
}

long file_compare(void) {
/*---------------------*/
long    offset;
short   equal,c1,c2;

offset=0L;
equal=TRUE;
do {
  c1=fgetc(file1);
  c2=fgetc(file2);
  offset++;
  if (c1==EOF || c2==EOF)
    break;
  if (c1!=c2)
    equal=FALSE;
} while (equal);
if (offset>0L)
  offset--;
if (equal && c1==EOF && c2==EOF)
  offset=-1L;
return(offset);
}

/* --- die groe Show ----------------------------------------- */

#define SHIFT   (RIGHT_SHIFT|LEFT_SHIFT)

void show_files(void) {
/*-------------------*/
short   ready,key;
long    offset;

which=1;
search_jump=' ';
search_n=0;
jump_byte=0x00;
mode='X';
init_display();
display1();
display2();

ready=FALSE;
while (!ready) {
  switch (key=sfc_getkey()) {
  case 'h':
  case 'H':
    display_help();
    break;
  case '1':
    which=1;
    sfc_goto_xy(11,1);
    break;
  case '2':
    which=2;
    sfc_goto_xy(11,13);
    break;
  case 'x':
  case 'X':
    mode='X';
    display1();
    display2();
    break;
  case 'a':
  case 'A':
    mode='A';
    display1();
    display2();
    break;
  case 'q':
  case 'Q':
    ready=TRUE;
    break;
  case 'c':
  case 'C':
    compare();
    break;
  case 's':
  case 'S':
    if (get_search()) {
      search_jump='S';
      search();
      }
    break;
  case 'j':
  case 'J':
    if (get_jump()) {
      search_jump='J';
      jump();
      }
    break;
  case 'r':
  case 'R':
    switch (search_jump) {
    case 'S':
      search();
      break;
    case 'J':
      jump();
      break;
    default:
      break;
      }
    break;
  case 'o':
  case 'O':
    offset=get_offset();
    if (offset<0L)
      break;
    if (which==1 || (sfc_kflags()&SHIFT)) {
      if (offset<=length1)
	set_absolute(1,offset);
      else
	set_absolute(1,length1);
      display1();
      }
    if (which==2 || (sfc_kflags()&SHIFT)) {
      if (offset<=length2)
	set_absolute(2,offset);
      else
	set_absolute(2,length2);
      display2();
      }
    break;
  case T_LEFT:
    if (which==1 || (sfc_kflags()&SHIFT)) {
      set_relative(1,-1L);
      display1();
      }
    if (which==2 || (sfc_kflags()&SHIFT)) {
      set_relative(2,-1L);
      display2();
      }
    break;
  case T_RIGHT:
    if (which==1 || (sfc_kflags()&SHIFT)) {
      set_relative(1,+1L);
      display1();
      }
    if (which==2 || (sfc_kflags()&SHIFT)) {
      set_relative(2,+1L);
      display2();
      }
    break;
  case T_UP:
    if (which==1 || (sfc_kflags()&SHIFT)) {
      set_relative_up(1);
      display1();
      }
    if (which==2 || (sfc_kflags()&SHIFT)) {
      set_relative_up(2);
      display2();
      }
    break;
  case T_DOWN:
    if (which==1 || (sfc_kflags()&SHIFT)) {
      set_relative_down(1);
      display1();
      }
    if (which==2 || (sfc_kflags()&SHIFT)) {
      set_relative_down(2);
      display2();
      }
    break;
  case T_PGUP:
    if (which==1 || (sfc_kflags()&SHIFT)) {
      set_relative_pgup(1);
      display1();
      }
    if (which==2 || (sfc_kflags()&SHIFT)) {
      set_relative_pgup(2);
      display2();
      }
    break;
  case T_PGDN:
    if (which==1 || (sfc_kflags()&SHIFT)) {
      set_relative_pgdown(1);
      display1();
      }
    if (which==2 || (sfc_kflags()&SHIFT)) {
      set_relative_pgdown(2);
      display2();
      }
    break;
  case T_HOME:
    if (which==1 || (sfc_kflags()&SHIFT)) {
      set_absolute(1,0L);
      display1();
      }
    if (which==2 || (sfc_kflags()&SHIFT)) {
      set_absolute(2,0L);
      display2();
      }
    break;
  case T_END:
    if (which==1 || (sfc_kflags()&SHIFT)) {
      set_absolute(1,length1);
      display1();
      }
    if (which==2 || (sfc_kflags()&SHIFT)) {
      set_absolute(2,length2);
      display2();
      }
    break;
  default:
    break;
    }
  if (key!=0)
    sfc_wait_nokey();
  }
exit_display();
}

/* --- Ausgabe ------------------------------------------------ */

void init_display(void) {
/*---------------------*/
sfc_cls();
display_title();
}

void exit_display(void) {
/*---------------------*/
sfc_goto_xy(0,24);
printf("\nfinished!\n");
}

void display_title(void) {
/*----------------------*/
sfc_line(0,"Smart FileCompare " VERSION "    "
	  "H = help   Q = quit");
if (which==1)
  sfc_goto_xy(11,1);
else
  sfc_goto_xy(11,13);
}

void display_title1(long offset) {
/*------------------------------*/
char    value[12];

sfc_invers(12,1,name1);
if (mode=='X')
  sprintf(value,"%11lX",offset);
else
  sprintf(value,"%11ld",offset);
sfc_invers(0,1,value);
}

void display1(void) {
/*-----------------*/
short   x,y,c,eol;
char    value[3],line[80+2];

display_title1(offset1);

fseek(file1,offset1,SEEK_SET);
if (mode=='X') {
  for (y=2; y<13; y++) {
    for (x=0; x<80; x++)
      line[x]=' ';
    line[80]='\0';
    for (x=0; x<16; x++) {
      c=fgetc(file1);
      if (c==EOF) {
	line[52+x]=' ';
	strcpy(value,"--");
      } else {
	line[52+x]=(char)c;
	sprintf(value,"%02X",c);
	}
      strncpy(&(line[3*x]),value,2);
      if (line[52+x]=='\0')
	line[52+x]=' ';
      }
    sfc_normal(0,y,line);
    }
} else {
  for (y=2; y<13; y++) {
    eol=FALSE;
    for (x=0; x<80; x++) {
      if (eol)
	line[x]=' ';
      else
	switch (c=fgetc(file1)) {
	case EOF:
	  eol=TRUE;
	case '\0':
	  line[x]=' ';
	  break;
	case LF:
	  eol=TRUE;
	default:
	  line[x]=(char)c;
	  break;
	  }
      }
    line[80]='\0';
    sfc_normal(0,y,line);
    }
  }
}

void display_title2(long offset) {
/*------------------------------*/
char    value[12];

sfc_invers(12,13,name2);
if (mode=='X')
  sprintf(value,"%11lX",offset);
else
  sprintf(value,"%11ld",offset);
sfc_invers(0,13,value);
}

void display2(void) {
/*-----------------*/
short   x,y,c,eol;
char    value[3],line[80+2];

display_title2(offset2);

fseek(file2,offset2,SEEK_SET);
if (mode=='X') {
  for (y=14; y<25; y++) {
    for (x=0; x<80; x++)
      line[x]=' ';
    line[80]='\0';
    for (x=0; x<16; x++) {
      c=fgetc(file2);
      if (c==EOF) {
	line[52+x]=' ';
	strcpy(value,"--");
      } else {
	line[52+x]=(char)c;
	sprintf(value,"%02X",c);
	}
      strncpy(&(line[3*x]),value,2);
      if (line[52+x]=='\0')
	line[52+x]=' ';
      }
    sfc_normal(0,y,line);
    }
} else {
  for (y=14; y<25; y++) {
    eol=FALSE;
    for (x=0; x<80; x++) {
      if (eol)
	line[x]=' ';
      else
	switch (c=fgetc(file2)) {
	case EOF:
	  eol=TRUE;
	case '\0':
	  line[x]=' ';
	  break;
	case LF:
	  eol=TRUE;
	default:
	  line[x]=(char)c;
	  break;
	  }
      }
    line[80]='\0';
    sfc_normal(0,y,line);
    }
  }
}

void display_help(void) {
/*---------------------*/
sfc_help_save();

sfc_normal(23, 6,"Smart FileCompare " VERSION " (" __DATE__ ")");
sfc_normal(13, 7,"moving:");
sfc_normal(22, 7,"cursor");
sfc_normal(31, 7,"left/right");
sfc_normal(51, 7,"up/down");
sfc_normal(31, 8,"page-up/page-down");
sfc_normal(51, 8,"home/end");
sfc_normal(22, 9,"1  focus at file 1");
sfc_normal(22,10,"2  focus at file 2");
sfc_normal(22,11,"O  sets offset");
sfc_normal(22,12,"shifted cursor works in both files");
sfc_normal(13,13,"display:");
sfc_normal(22,13,"A  as text");
sfc_normal(22,14,"X  as hexbytes");
sfc_normal(13,15,"work:");
sfc_normal(22,15,"C  starts comparing");
sfc_normal(22,16,"S  searches for bytes");
sfc_normal(22,17,"J  jumps to another byte");
sfc_normal(22,18,"R  repeat search/jump");
sfc_normal(22,19,"Q  quits program");
sfc_normal(52,19,"press any key...");
sfc_goto_xy(68,19);

while (!sfc_kbhit());
sfc_wait_nokey();

sfc_help_restore();
}

/* --- Eingabe ------------------------------------------------ */

long get_offset(void) {
/*-------------------*/
short   ready,update,key;
long    value;
char    line[11];

value=0;
if (mode=='X') {
  sfc_line(0,"offset (hex): ######## 0..9/A..F/BS/CR/ESC");
  sfc_goto_xy(21,0);
} else {
  sfc_line(0,"offset (dec): ########## 0..9/BS/CR/ESC");
  sfc_goto_xy(23,0);
  }

ready=0;
update=TRUE;
do {
  if (update) {
    if (mode=='X')
      sprintf(line,"%8lX",value);
    else
      sprintf(line,"%10ld",value);
    sfc_normal(14,0,line);
    update=FALSE;
    }
  switch (key=sfc_getkey()) {
  case T_BACKSPACE:
    if (mode=='X')
      value/=16;
    else
      value/=10;
    update=TRUE;
    break;
  case T_RETURN:
    ready=1;
    break;
  case T_ESC:
    ready=2;
    break;
  default:
    if (mode=='X') {
      if (value<=LONG_MAX/16 && isxdigit(key)) {
	value*=16;
	if (isdigit(key))
	  value+=key-'0';
	else
	  value+=toupper(key)-'A'+10;
	update=TRUE;
	}
    } else {
      if (value<=LONG_MAX/10 && isdigit(key)) {
	value*=10;
	value+=key-'0';
	update=TRUE;
	}
      }
    break;
    }
} while (ready==0);

if (ready!=1)
  value=-1L;
display_title();
return(value);
}

short get_search(void) {
/*--------------------*/
short           ready,update,key,n,i;
unsigned char   value[4+1],line[4*3+1],*p;

n=search_n;
for (i=0; i<n; i++) {
  value[i]=(unsigned char)search_byte[i];
  if (mode!='X' && value[i]<' ')
    value[i]=' ';
  }
value[n]='\0';

if (mode=='X') {
  sfc_line(0,"search (hex): ##,##,##,## 0..9/A..F/,/BS/CR/ESC");
  sfc_goto_xy(24,0);
  p=line;
  for (i=n; i<4; i++) {
    *p++=ASCII_NONE;
    *p++=ASCII_NONE;
    *p++=',';
    }
  for (i=0; i<n; i++) {
    sprintf((char*)p,"%02X,",value[i]);
    p+=3;
    }
  line[11]='\0';
  sfc_normal(14,0,(char *)line);
} else {
  sfc_line(0,"search (ascii): #### char/BS/CR/ESC");
  sfc_goto_xy(19,0);
  sprintf((char*)line,"%4s",value);
  for (i=n; i<4; i++)
    line[3-i]=ASCII_NONE;
  sfc_normal(16,0,(char *)line);
  }

ready=0;
update=TRUE;
do {
  if (update) {
    if (mode=='X') {
      p=line;
      for (i=n; i<4; i++) {
	*p++=ASCII_NONE;
	*p++=ASCII_NONE;
	*p++=',';
	}
      for (i=0; i<n; i++) {
	sprintf((char*)p,"%02X,",value[i]);
	p+=3;
	}
      line[11]='\0';
      sfc_normal(14,0,(char *)line);
    } else {
      sprintf((char*)line,"%4s",value);
      for (i=n; i<4; i++)
	line[3-i]=ASCII_NONE;
      sfc_normal(16,0,(char *)line);
      }
    update=FALSE;
    }
  switch (key=sfc_getkey()) {
  case T_BACKSPACE:
    if (n>0)
      n--;
    value[n]='\0';
    update=TRUE;
    break;
  case T_RETURN:
    ready=1;
    break;
  case T_ESC:
    ready=2;
    break;
  case ',':
    if (mode=='X') {
      if (n==4) {
	n--;
	for (i=0; i<n; i++)
	  value[i]=value[i+1];
	}
      value[n++]=0x00;
      update=TRUE;
      }
    break;
  default:
    if (mode=='X') {
      if (isxdigit(key)) {
	value[n-1]*=16;
	if (isdigit(key))
	  value[n-1]+=(unsigned char)key-'0';
	else
	  value[n-1]+=(unsigned char)toupper(key)-'A'+10;
	update=TRUE;
	}
    } else {
      if (key>=' ' && key<=0xFF) {
	if (n==4) {
	  n--;
	  for (i=0; i<n; i++)
	    value[i]=value[i+1];
	  }
	value[n++]=(unsigned char)key;
	value[n]='\0';
	update=TRUE;
	}
      }
    break;
    }
} while (ready==0);

if (ready==1) {
  for (i=0; i<n; i++)
    search_byte[i]=value[i];
  search_n=n;
  }
display_title();
return(ready==1 && search_n>0);
}

short get_jump(void) {
/*------------------*/
short   ready,update,key,value;
char    line[4];

value=jump_byte;
if (mode!='X' && value<' ')
  value=' ';

if (mode=='X') {
  sfc_line(0,"jump over (hex): ## 0..9/A..F/CR/ESC");
  sfc_goto_xy(18,0);
} else {
  sfc_line(0,"jump over (ascii): # char/CR/ESC");
  sfc_goto_xy(19,0);
  }

ready=0;
update=TRUE;
do {
  if (update) {
    if (mode=='X') {
      sprintf(line,"%02X",value);
      sfc_normal(17,0,line);
    } else {
      sprintf(line,"%c",value);
      sfc_normal(19,0,line);
      }
    update=FALSE;
    }
  switch (key=sfc_getkey()) {
  case T_RETURN:
    ready=1;
    break;
  case T_ESC:
    ready=2;
    break;
  default:
    if (mode=='X') {
      if (isxdigit(key)) {
	value*=16;
	if (isdigit(key))
	  value+=key-'0';
	else
	  value+=toupper(key)-'A'+10;
	value&=0xFF;
	update=TRUE;
	}
    } else {
      if (key>=' ' && key<=0xFF) {
	value=key;
	update=TRUE;
	}
      }
    break;
    }
} while (ready==0);

if (ready==1)
  jump_byte=value;
display_title();
return(ready==1);
}

/* --- Vergleichen und Suchen --------------------------------- */

void compare(void) {
/*----------------*/
short   c1,c2,k;

sfc_line(0,"comparing... (any key for abort)");
fseek(file1,offset1,SEEK_SET);
fseek(file2,offset2,SEEK_SET);
k=0;
do {
  offset1++;
  offset2++;
  c1=fgetc(file1);
  c2=fgetc(file2);
  if (c1==EOF || c2==EOF)
    break;
  if (++k==KEY_SLICE) {
    k=0;
    display_title1(offset1);
    display_title2(offset2);
    if (sfc_kbhit())
      break;
    }
} while (c1==c2);
offset1--;
offset2--;
display1();
display2();
display_title();
}

void search(void) {
/*---------------*/
short   ready,i,c,k;
long    difference;

sfc_line(0,"searching... (any key for abort)");
if (which==1)
  fseek(file1,offset1+1L,SEEK_SET);
else
  fseek(file2,offset2+1L,SEEK_SET);
ready=FALSE;
difference=0L;
i=0;
k=0;
do {
  if (which==1)
    c=fgetc(file1);
  else
    c=fgetc(file2);
  difference++;
  if (c==EOF) {
    difference=0L;
    ready=TRUE;
  } else if (c==search_byte[i]) {
    if (++i==search_n) {
      difference-=i-1;
      ready=TRUE;
      }
  } else
    i=0;
  if (++k==KEY_SLICE) {
    k=0;
    if (which==1)
      display_title1(offset1+difference);
    else
      display_title2(offset2+difference);
    if (sfc_kbhit()) {
      difference=0L;
      ready=TRUE;
      }
    }
} while (!ready);
if (difference>0L) {
  set_relative(which,difference);
  if (which==1)
    display1();
  else
    display2();
} else {
  if (which==1)
    display_title1(offset1);
  else
    display_title2(offset2);
  }
display_title();
}

void jump(void) {
/*-------------*/
short   ready,c,k;
long    difference;

sfc_line(0,"jumping... (any key for abort)");
if (which==1)
  fseek(file1,offset1+1L,SEEK_SET);
else
  fseek(file2,offset2+1L,SEEK_SET);
ready=FALSE;
difference=0L;
k=0;
do {
  if (which==1)
    c=fgetc(file1);
  else
    c=fgetc(file2);
  difference++;
  if (c==EOF) {
    difference=0L;
    ready=TRUE;
  } else if (c!=jump_byte)
    ready=TRUE;
  if (++k==KEY_SLICE) {
    k=0;
    if (which==1)
      display_title1(offset1+difference);
    else
      display_title2(offset2+difference);
    if (sfc_kbhit()) {
      difference=0L;
      ready=TRUE;
      }
    }
} while (!ready);
if (difference>0L) {
  set_relative(which,difference);
  if (which==1)
    display1();
  else
    display2();
} else {
  if (which==1)
    display_title1(offset1);
  else
    display_title2(offset2);
  }
display_title();
}

/* --- Offset verndern --------------------------------------- */

void set_relative(short who,long difference) {
/*------------------------------------------*/
if (who==1) {
  offset1+=difference;
  fseek(file1,offset1,SEEK_SET);
  if (offset1>length1 || ftell(file1)!=offset1)
    offset1-=difference;
} else {
  offset2+=difference;
  fseek(file2,offset2,SEEK_SET);
  if (offset2>length2 || ftell(file2)!=offset2)
    offset2-=difference;
  }
}

void set_absolute(short who,long offset) {
/*--------------------------------------*/
if (who==1) {
  fseek(file1,offset,SEEK_SET);
  if (offset<=length1 || ftell(file1)==offset)
    offset1=offset;
} else {
  fseek(file2,offset,SEEK_SET);
  if (offset<=length2 || ftell(file2)==offset)
    offset2=offset;
  }
}

short frgetc(FILE *file) {
/*----------------------*/
long    offset;
short   c;

offset=ftell(file);
fseek(file,-1L,SEEK_CUR);
if (ftell(file)!=offset-1L)
  c=EOF;
else {
  c=fgetc(file);
  fseek(file,-1L,SEEK_CUR);
  }
return(c);
}

void set_relative_up(short who) {
/*-----------------------------*/
short   i,eol,c;

if (mode=='X') {
  for (i=0; i<16; i++)
    set_relative(who,-1L);
} else {
  if (who==1)
    fseek(file1,offset1,SEEK_SET);
  else
    fseek(file2,offset2,SEEK_SET);
  i=0;
  eol=FALSE;
  while (i<=80 && !eol) {
    if (who==1)
      c=frgetc(file1);
    else
      c=frgetc(file2);
    i++;
    switch (c) {
    case LF:
      if (i!=1)
	eol=TRUE;
      break;
    case EOF:
      eol=TRUE;
      break;
    default:
      break;
      }
    }
  i--;
  set_relative(who,-(long)i);
  }
}

void set_relative_down(short who) {
/*-------------------------------*/
short   i,eol,c;

if (mode=='X') {
  for (i=0; i<16; i++)
    set_relative(who,+1L);
} else {
  if (who==1)
    fseek(file1,offset1,SEEK_SET);
  else
    fseek(file2,offset2,SEEK_SET);
  i=0;
  eol=FALSE;
  while (i<80 && !eol) {
    if (who==1)
      c=fgetc(file1);
    else
      c=fgetc(file2);
    i++;
    switch (c) {
    case LF:
      eol=TRUE;
      break;
    case EOF:
      i--;
      eol=TRUE;
      break;
    default:
      break;
      }
    }
  set_relative(who,(long)i);
  }
}

void set_relative_pgup(short who) {
/*-------------------------------*/
short   i;

if (mode=='X')
  for (i=0; i<160; i++)
    set_relative(who,-1L);
else
  for (i=0; i<10; i++)
    set_relative_up(who);
}

void set_relative_pgdown(short who) {
/*---------------------------------*/
short   i;

if (mode=='X')
  for (i=0; i<160; i++)
    set_relative(who,+1L);
else
  for (i=0; i<10; i++)
    set_relative_down(who);
}

/* --- Ende --------------------------------------------------- */
